/*********************************************************************
   Desktop Station Core Library for Arduino

   Copyright (C) 2024 Desktop Station Co.,Ltd.
   Not use for commercial use
*/

#define DEBUG

#include <arduino.h>
#include "hardware/adc.h" 
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM.h"
#include "board_pins.h"


#define THRESHOLD_CVREAD 2
#define THRESHOLD_CURRENT 20  /*  */
#define COUNT_RETRY		2
#define SIZE_CVBITS   16


static DS_ITEM gList[ITEM_LISTSIZE];/**< 送信管理用リスト */
static uint8_t gScanIndex = 0;		/**< スキャンインデックス（リスト走査） */
static uint8_t gPriorityIndex = 0;		/**< 優先度インデックス */
static uint8_t gSwitchScan = 0;
static boolean gPoweron = false;
unsigned long gPreviousTimeRC = 0;

word gCVAckMaxVal;
word gCVAckMinVal;
word gCVAckAdcOffset;


//task
repeating_timer_t timer_1ms_cvack_task; // timer変数の定義


//定周期処理
bool timer_callback_cvack_1ms(repeating_timer_t *rt);

//---------------------------------------------------------------------
// CV読み出し時Ack検出用定周期タスク(1ms周期) 
//---------------------------------------------------------------------

bool timer_callback_cvack_1ms(repeating_timer_t *rt)
{
	//CV Ack検出シーケンス
	
	//ADCの選択(電流検出を選択しておく)
	adc_select_input(PIN_CURRENT_A2);

	//ADC読み出し
	long aCurrent = adc_read();
	
	if( aCurrent > gCVAckAdcOffset)
	{
		aCurrent = aCurrent - gCVAckAdcOffset;
	}
	else
	{
		aCurrent = 0;
	}
	
	//極大極小チェック
	if( gCVAckMaxVal < aCurrent)
	{
		gCVAckMaxVal = aCurrent;
	}
	
	if( gCVAckMinVal > aCurrent)
	{
		gCVAckMinVal = aCurrent;
	}
	
	return true;
}



DSCoreLib::DSCoreLib()
{

	index_send = 0;
	state_cvscan = 0;
	count_cvscan_state = 0;
	data_end32_last_send = 0;
	gTxBuf[0] = 0;

	gPoweron = false;
	gPriorityIndex = 0;
	gSwitchScan = 0;
	
	flag_send_twice = 0;
	len_send_twice = 0;
	
	gThresholdCurrent = THRESHOLD_CURRENT;
	
	gRetryCount = COUNT_RETRY;

	//前回送信時のデータ
	len_last_sendTx = 0;
	len_last_sendBits = 0;	
	
	gCutOut = 0;
	
	//cv関連
	offset_adc_current = 0;

}


uint8_t DSCoreLib::MAIN_scanList(DS_ITEM *inList, uint8_t inScanIndex)
{
	uint8_t aRetIndex = inScanIndex;

	if ( inScanIndex >= ITEM_LISTSIZE)
	{
		return 255;
	}

	/* パケットを送信する */
	if ( inList[inScanIndex].mType == ITEMTYPE_DCC )
	{
		//Send DCC pulse
		sendPackets(inList[inScanIndex].mDatas, inList[inScanIndex].mLen, DCCPREAMBLE_DEFAULT);
		
		/* 高優先度・回数制限のあるパケットは二回送る */
		if ((inList[inScanIndex].mPriority  > 0 ) || ( inList[inScanIndex].mCnt > 0))
		{
			flag_send_twice = 1;
		}
		else
		{
			flag_send_twice = 0;
		}
		
		if ( inList[inScanIndex].mPriority > 0)
		{
			inList[inScanIndex].mPriority--;
		}
	}
	else if ( inList[inScanIndex].mType == ITEMTYPE_UNKNOWN )
	{
		aRetIndex = 255;
	}
	else
	{
		aRetIndex = 255;
	}

	/* カウント値が入っている指令は時間が来たら削除する */
	if ( (inList[inScanIndex].mCnt > 0) && (aRetIndex != 255))
	{
		inList[inScanIndex].mCnt--;
		
		if ( inList[inScanIndex].mCnt <= 0)
		{
			DSList_DeleteItem(inList, inScanIndex);
		}
	}
	
	return aRetIndex;
}

uint8_t DSCoreLib::Main_scanFunction(DS_ITEM *inList, uint8_t inScanIndex)
{
	uint8_t aRetIndex = inScanIndex;

	if ( inScanIndex >= ITEM_LISTSIZE)
	{
		return 255;
	}

	if ( inList[inScanIndex].mType == ITEMTYPE_DCC )
	{
		Main_generateDCCFncPacket(inList, inScanIndex, (inList[inScanIndex].mPriority > 0) ? 1 : 0);
	}
	else
	{
		aRetIndex = 255;
	}
	
	return aRetIndex;
}


uint8_t DSCoreLib::Main_generateDCCFncPacket(DS_ITEM *inList, uint8_t inScanIndex, uint8_t inSendTwice)
{

  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aOutputLen = 0;
  uint64_t aFunctionBuf;
  uint8_t aOffset = 0;


  /* パケットのアドレス部分は確定するのでセットする */
  aPackets[0] = inList[inScanIndex].mDatas[0];

  if ( aPackets[0] == 0xFF)
  {
    /* アイドルor停止パケットなので終了する */
    return 0;
  }
  else if ( aPackets[0] == 0x00)
  {
  	if( (inList[inScanIndex].mDatas[1] == 0))
  	{
		/* リセットパケットなので終了する */
		return 0;
  	}
  }
  else if ((aPackets[0] >= 128) && (aPackets[0] <= 191))
  {
    /* ポイント・信号・アクセサリ用なので終了する */
    return 0;
  }
  else if ((aPackets[0] >= 232) && (aPackets[0] <= 254))
  {
    /* リザーブ領域なので終了する */
    return 0;
  }
  else if ( (aPackets[0] >= 192) && (aPackets[0] <= 231))
  {
    /* 拡張ロコパケット */
    aPackets[1] = inList[inScanIndex].mDatas[1];
    aOffset = 1;
  }
  else if ((aPackets[0] >= 1) && (aPackets[0] <= 127))
  {
    /* ベーシックロコパケット */
    aOffset = 0;
  }
  else
  {
    /* 不明なパケット */
    return 2;
  }

  /* decode function  */

  aFunctionBuf = inList[inScanIndex].mFunctionBuf;

  /* ファンクションの書き換えとDCCのパケット生成 */
  aPackets[1 + aOffset] = DCC_decodeFunctionRaw(inList[inScanIndex].mFuncCounter, &aFunctionBuf);

  /* F13-F20, F21-28の拡張ファンクション用（ファンクション用にサイズが増える） */
  switch (aPackets[1 + aOffset])
  {
    case 0b11011110:
      aPackets[2 + aOffset] = ((aFunctionBuf >> 13) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;

    case 0b11011111:
      aPackets[2 + aOffset] = ((aFunctionBuf >> 21) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;

    case 0b11011000:/* 拡張アドレス(F29-F36) */
      aPackets[2 + aOffset] = ((aFunctionBuf >> 29) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;

    case 0b11011001:/* 拡張アドレス(F37-F44) */
      aPackets[2 + aOffset] = ((aFunctionBuf >> 37) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;
  	
    case 0b11011010:/* 拡張アドレス(F45-F52) */
      aPackets[2 + aOffset] = ((aFunctionBuf >> 45) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;
  	
    case 0b11011011:/* 拡張アドレス(F53-F60) */
      aPackets[2 + aOffset] = ((aFunctionBuf >> 53) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;

  	case 0b11011100:/* 拡張アドレス(F61-F68) */
      aPackets[2 + aOffset] = ((aFunctionBuf >> 61) & 0xFF);
      aOutputLen = 3 + aOffset;
      break;



    default:
      aOutputLen = 2 + aOffset;
      break;
  }

	/* Output Packet */
	sendPackets(aPackets, aOutputLen, DCCPREAMBLE_DEFAULT);

	/* Send 2 packets */
	if( inSendTwice > 0)
	{
		flag_send_twice = 1;
	}
	else
	{
		flag_send_twice = 0;
	}


  /* Increment for next packet generation */
  inList[inScanIndex].mFuncCounter = inList[inScanIndex].mFuncCounter + 6;

  if ( inList[inScanIndex].mFuncCounter > 36)
  {
    inList[inScanIndex].mFuncCounter = 0;
  }

  return 1;
}


uint8_t DSCoreLib::Main_decodeLocDirection(DS_ITEM *inList, word address, uint8_t direction, uint8_t *outType, uint8_t *outPackets, uint8_t *outLen)
{
  uint8_t aAddr1, aAddr2;
  uint8_t aLen;
  uint8_t aType;
  uint8_t aDirection; /* 0:FWD, 1:REV */
  uint8_t aIndex;
  uint8_t aSpeedRecoveryPacket;
  uint8_t aSpeed;

  /* decode loc address */
  Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);

  if( aLen == 0)
  {
  	return 0;
  }


  /* パケットのアドレス部分は確定するのでセットする */
  outPackets[0] = aAddr1;

  if ( aLen >= 2)
  {
    outPackets[1] = aAddr2;
  }

  *outType = aType;

  /* 前回のファンクションデータが必要なため同じアドレスを探索 */
  aIndex = DSList_FindItem(inList, aType, CMD_SPEED, aLen, 0, outPackets);

  if ( aIndex != 255)
  {
    aDirection = inList[aIndex].mDirection;
  }
  else
  {
    return 255;
  }

  if ( aDirection != direction)
  {

	aSpeedRecoveryPacket = 0;

    /* 異なる場合はリスト側を書き換える */
    inList[aIndex].mDirection = direction;
    inList[aIndex].mReverse = aSpeedRecoveryPacket;

    return 1;
  }
  else
  {
    /* 同じ場合は書き換える必要なし */
    return 0;
  }

}

void DSCoreLib::Main_decodeLocFunction(DS_ITEM *inList, word address, uint8_t function, uint8_t power, uint8_t *outType, uint8_t *outPackets, uint8_t *outLen, uint64_t *ioFuncBuf)
{
  uint8_t aAddr1, aAddr2;
  uint8_t aLen;
  uint8_t aType;
  uint32_t aFunctionBuf;
  uint8_t aIndex;
  uint8_t aOffset = 0;

  /* decode loc address */
  Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);

  if( aLen == 0)
  {
  	return;
  }

  /* パケットのアドレス部分は確定するのでセットする */
  *outType = aType;
  outPackets[0] = aAddr1;

  if ( aLen >= 2)
  {
    outPackets[1] = aAddr2;
    aOffset = 1;
  }
  else
  {
    aOffset = 0;
  }

  /* 前回のファンクションデータが必要なため同じアドレスを探索 */
  aIndex = DSList_FindItem(inList, aType, CMD_SPEED, aLen, 0, outPackets);


	
  if ( aIndex != 255)
  {
    aFunctionBuf = inList[aIndex].mFunctionBuf;
  }
  else
  {
    aFunctionBuf = 0;
  }

  /* Function buffer update  */
	if( (function > 0) && ( function <= 61))
	{
		if ( power == 0)
		{
			*ioFuncBuf = aFunctionBuf & (~(1ULL << (function - 1)));
		}
		else
		{
			*ioFuncBuf = aFunctionBuf | (1ULL << (function - 1));
		}
	}

	/* Generate dammy packet for initial registration */
	if( aIndex == 255)
	{
		switch ( aType)
		{
		case ITEMTYPE_DCC:
		  /* 速度ゼロのダミー */
		  outPackets[1 + aOffset] = 0b01100000;
		  *outLen = 2 + aOffset;
		  break;
		}
	}

}

void DSCoreLib::Main_decodeLocSpeed(DS_ITEM *inList, word address, short speed, uint8_t *outType, uint8_t *outPackets, uint8_t *outLen, uint8_t inSpeedStep)
{
  uint8_t aAddr1, aAddr2;
  uint8_t aLen;
  uint8_t aType;
  uint8_t aDirection;
  uint8_t aIndex;
  uint8_t aSpeed;
  uint8_t aOffset;

  /* decode loc address */
  Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);

  if( aLen == 0)
  {
  	return;
  }
	
  /* パケットのアドレス部分は確定するのでセットする */
  *outType = aType;
  outPackets[0] = aAddr1;

  if ( aLen >= 2)
  {
    outPackets[1] = aAddr2;
  }

  /* 前回のファンクションデータが必要なため同じアドレスを探索 */
  aIndex = DSList_FindItem(inList, aType, CMD_SPEED, aLen, 0, outPackets);

  if ( aIndex != 255)
  {
    aDirection = inList[aIndex].mDirection;
  }
  else
  {
    aDirection = 0;
  }


	
  /* Speedstepのセット */
  if ( inSpeedStep != 0xFF)
  {
    inList[aIndex].mSpeedStep = inSpeedStep;
  }

  /* decode speed  */

  switch ( aType)
  {
    case ITEMTYPE_DCC:

      aOffset = (aLen >= 2) ? 1 : 0;

      if (inList[aIndex].mSpeedStep == SPEEDSTEP_DCC28)
      {
        /* 28step速度 */
        outPackets[1 + aOffset] = DCC_decodeSpeedStep28(speed, aDirection);
        *outLen = 2 + aOffset;
      }
      else if (inList[aIndex].mSpeedStep == SPEEDSTEP_DCC127)
      {
        /* 拡張127step速度 */
        outPackets[1 + aOffset] = 0b00111111;
        outPackets[2 + aOffset] = DCC_decodeSpeedStep127(speed, aDirection);
        *outLen = 3 + aOffset;
      }
      else if (inList[aIndex].mSpeedStep == SPEEDSTEP_DCC14)
      {
        /* 14step速度 */
        outPackets[1 + aOffset] = DCC_decodeSpeedStep14(speed, aDirection, (uint8_t)(inList[aIndex].mFunctionBuf & 0b1));
        *outLen = 2 + aOffset;
      }
      else
      {
        outPackets[1 + aOffset] = 0;
        *outLen = 2 + aOffset;
      }

      break;
  }


}

void DSCoreLib::Main_decodeAccessory(word address, uint8_t power, uint8_t *outType, uint8_t *outPackets, uint8_t *outLen)
{
  uint8_t aAddr1, aAddr2;
  uint8_t aLen;
  uint8_t aType;
  uint8_t aPower;
  uint8_t aProtocolType = DS_GetLocIDProtocol((uint8_t)(address >> 8));

  /* decode position  */

  switch ( aProtocolType)
  {
    case PROTOCOL_DCC_ACC:
      /* decode acc address (basic 9bit accessory mode) */
      Main_decodeAddress( address, &aType, &aAddr1, &aAddr2, &aLen);

      /* Power: 0 Straight(Green), 1 Diverging(Red) */

      *outType = ITEMTYPE_DCC;
      outPackets[0] = aAddr1;
      outPackets[1] = aAddr2 | 0b1000 | (power & 0b1) | ((((address - ADDR_ACC_DCC) % 4) & 0b11)  << 1);//1AAACDDD
      *outLen = 2;

      break;
  }


}


void DSCoreLib::Main_decodeAddress( word address, uint8_t *outpType, uint8_t *outpAddr1, uint8_t *outpAddr2, uint8_t *outpLen)
{
  word aTempAddr;
  uint8_t aType = DS_GetLocIDProtocol((uint8_t)(address >> 8));

  switch (aType)
  {
    case PROTOCOL_DCC:
      *outpType = ITEMTYPE_DCC;

      /* offset DCC address to convert CAN address to DCC address */
      aTempAddr = address - 0xC000;

      if (( aTempAddr > 0) && ( aTempAddr <= 127))
      {
        /* 0-127 address */
        *outpAddr1 = (uint8_t)(aTempAddr & 0b01111111);
        *outpAddr2 = 0;
        *outpLen = 1;
      }
      else if (( aTempAddr > 127) && ( aTempAddr < 10239))
      {
        /* DCC 14bit address range is 0b11000000-0b00000000(49152, addr.0) to 0b11100111-0b11111111(59391) */
        /* 128-10239 address, but 0-127 long address doesn't use */
        aTempAddr = aTempAddr + 49152;
        *outpAddr1 = (uint8_t)((aTempAddr >> 8) & 0xFF);
        *outpAddr2 = (uint8_t)(aTempAddr & 0xFF);
        *outpLen = 2;
      }
      else if (aTempAddr == 10240)
	{
		//address=59392d(0xE800)
		/* Short Address Broadcast */
        *outpAddr1 = 0;//Broadcast
        *outpAddr2 = 0;
        *outpLen = 1;
	}
      else
      {
        /* out of 9bit/14bit locomotive address range */
        *outpLen = 0;

      }

      break;

    case PROTOCOL_DCC_ACC:
      *outpType = ITEMTYPE_DCC;

      /* offset DCC address to convert CAN address to DCC address */
      aTempAddr = ((address - ADDR_ACC_DCC) >> 2) + 1;

      if ( aTempAddr < 512)
      {
        *outpAddr1 = (uint8_t)(0b10000000 | (aTempAddr & 0b00111111));//10AAAAAA
        *outpAddr2 = (uint8_t)(0b10000000 | ((((~aTempAddr) & 0b111000000) >> 2) & 0xFF));//1AAACDDD
        *outpLen = 2;
      }
      else
      {
        /* out of 9bit accessory address range */
        //Extended Accessory Decoder Control Packet Format
        *outpLen = 0;

      }

      break;
    default:
      *outpLen = 0;
      break;
  }

}

void DSCoreLib::SendReset()
{

  /* 全DCCデコーダリセット処理 */
	for( uint8_t i = 0; i < 10; i++)
	{
		sendInitialPacket(DCCPREAMBLE_DEFAULT);
	}
}


void DSCoreLib::Init()
{
	
	InitPins();
	
	//電流検出用
	adc_gpio_init(PIN_CURRENT);
	
	
	/* 初期化 */
	Clear();

}

void DSCoreLib::InitPins()
{
	
	gpio_init(PIN_DCC_A);
	gpio_init(PIN_DCC_B);
	gpio_set_dir(PIN_DCC_A, GPIO_OUT);
	gpio_set_dir(PIN_DCC_B, GPIO_OUT);
	
	DS_OutputOff();
	
}


void DSCoreLib::Clear()
{
	/* リストスキャンを初期化 */
	gScanIndex = 0;
	gSwitchScan = 0;
	DSList_Clear(gList);
	
	
	/* バッファもクリア */
	ClearPacketStack();
}

uint8_t DSCoreLib::Scan()
{
	uint8_t aIdlePacket[2];
	uint8_t aPacketType = 0;
	bool aSendIdle = false;
	uint8_t aPackets0 = 0;

	if ( gSwitchScan == 0)
	{
		/* リストのスキャン */
		if ( MAIN_scanList(gList, gScanIndex) != 255)
		{
			/* 次のパケットへ */
			gScanIndex++;
	
		}
		else
		{
			/* リセット */
			gScanIndex = 0;

			/* ファンクション送信用にスイッチ */
			gSwitchScan = 1;

			/* DCC Idle */
			aSendIdle = true;
			
		}
	}
	else
	{
		//ファンクションが送信できるIndexまで進む
		while( gScanIndex < ITEM_LISTSIZE)
		{
			//ロコ用のパケットを探索する
			aPackets0 = gList[gScanIndex].mDatas[0];

			if ( ((aPackets0 >= 192) && (aPackets0 <= 231)) || ((aPackets0 >= 1) && (aPackets0 <= 127)))
			{
				if ( Main_scanFunction(gList, gScanIndex) == 255)
				{
					/* リセット */
					gScanIndex = 0;
					gSwitchScan = 0;

					if ( MAIN_scanList(gList, gScanIndex) != 255)
					{
						/* 次のパケットへ */
						gScanIndex++;
					}
					else
					{			
						/* DCC Idle */
						aSendIdle = true;				
					}
				}
				else
				{
					gScanIndex++;
				}
				break;
			}
			else if( aPackets0 == 0)
			{
				/* リセット */
				gScanIndex = 0;
				gSwitchScan = 0;

				if ( MAIN_scanList(gList, gScanIndex) != 255)
				{
					/* 次のパケットへ */
					gScanIndex++;
				}
				else
				{			
					/* DCC Idle */
					aSendIdle = true;				
				}
				break;				
			}

			gScanIndex++;
		}
	}

	if( aSendIdle == true)
	{
		/* DCC Idle */
		aIdlePacket[0] = 0b11111111;
		aIdlePacket[1] = 0b00000000;

		// 8*3 + 4 + PREAMBLE = 32 -28 = 4
		
		//Send DCC pulse
		sendPackets(aIdlePacket, 2, DCCPREAMBLE_DEFAULT);

		//アイドル
		aPacketType = 1;		
	}
	
	
	//DEBUG
	/*
	Serial.print("Packet: ");
	Serial.print(gScanIndex);
	Serial.print("/");
	Serial.print(gSwitchScan);
	Serial.print(" ");

	word aBitNum = 0;
	
	for(int i = 0; i < len_last_sendTx + 1; i++)
	{
		for( int j = 0; j < 32; j++)
		{

			if( aBitNum < len_last_sendBits)
			{
				Serial.print((gTxBuf[i] >> j) & 1);
			}
			else
			{
				Serial.print("x");
			}

			aBitNum++;
		}
		
		Serial.print(" ");
	}
	
	
	Serial.println();
	*/
	
	return aPacketType;
}


boolean DSCoreLib::IsPower()
{
  return gPoweron;
}

boolean DSCoreLib::SetPower(boolean power)
{

  if (power)
  {
    /* 線路に電源投入 */
    DS_Power(1);

    gPoweron = true;
  }
  else
  {
    DS_Power(0);

    gPoweron = false;
  }

  return true;
}


boolean DSCoreLib::SetLocoSpeed(word address, int inSpeed)
{
  return SetLocoSpeedEx( address, inSpeed, 0xFF);
}


boolean DSCoreLib::SetLocoSpeedEx(word address, int inSpeed, int inProtcol)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen = 0;
  uint8_t aType = 0;

  /* アドレスデコード処理 */
  Main_decodeLocSpeed(gList, address, inSpeed, &aType, aPackets, &aLen, inProtcol);

  /* 管理リストをアップデート */
  DSList_UpdateItem(gList, aType, CMD_SPEED, aLen, 0, aPackets);

  return true;
}

boolean DSCoreLib::SetLocoFunction(word address, unsigned char inFunction, unsigned char inPower)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen = 0;
  uint8_t aType = 0;
  uint64_t aFuncBuf = 0;
  
	if(inFunction > 61)
	{
		return false;
	}


  /* ファンクション指令デコード */
  Main_decodeLocFunction(gList, address, inFunction, inPower, &aType, aPackets, &aLen, &aFuncBuf);

  /* ファンクションはバッファのみ上書き(ファンクションパケットはスキャン時に自動生成) */
  DSList_UpdateItem_OWFunction(gList, aType, CMD_SPEED, aLen, aFuncBuf, inFunction, aPackets);

  return true;
}

boolean DSCoreLib::SetLocoFunctionRaw(word address, uint64_t inFunctions)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen = 0;
  uint8_t aType = 0;
  uint64_t aFuncBuf = 0;


  /* ファンクション指令デコード */
  Main_decodeLocFunction(gList, address, 0, inFunctions & 1, &aType, aPackets, &aLen, &aFuncBuf);

  /* ファンクションはバッファのみ上書き(ファンクションパケットはスキャン時に自動生成) */
  DSList_UpdateItem_OWFunction(gList, aType, CMD_SPEED, aLen, inFunctions, 0, aPackets);

  return true;
}


boolean DSCoreLib::SetLocoDirection(word address, unsigned char inDirection)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen = 0;
  uint8_t aType = 0;
  uint8_t aDir = inDirection;

  /* アドレスデコード処理,リストと指令が異なる場合は停止させる */
  switch ( Main_decodeLocDirection(gList, address, aDir, &aType, aPackets, &aLen) )
  {
    case 1:
      /* 速度をゼロにして止める */
      Main_decodeLocSpeed(gList, address, 0, &aType, aPackets, &aLen, 0xFF);
      DSList_UpdateItem(gList, aType, CMD_SPEED, aLen, 0, aPackets);
      break;

    case 255:
      /* 新規登録処理 */
      Main_decodeLocSpeed(gList, address, 0, &aType, aPackets, &aLen, 0xFF);
      /* 登録リストに入っていないものは、リストに新規登録（進行方向を格納するため） */
      DSList_NewItem(gList, aType, CMD_SPEED, aLen, 0, aDir, aPackets);
      break;

    default:
      break;
  }

  return true;
}


boolean DSCoreLib::SetTurnout(word address, uint8_t inSwitch)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen = 0;
  uint8_t aType = 0;

  /* アクセサリ指令デコード */
  Main_decodeAccessory(address, inSwitch, &aType, aPackets, &aLen);
	
  if( aLen == 0)
  {
  	return true;
  }


  /* 管理リストをアップデート */
  DSList_UpdateItem(gList, aType, CMD_ACCESSORY, aLen, 2, aPackets);

  return true;
}

word DSCoreLib::GetLocIDProtocol(uint8_t address)
{
  if ( address < 0x04)
  {
    return ADDR_MM2;
  }
  else if ( (address >= 0x30) && (address <= 0x33))
  {
    return ADDR_ACC_MM2;
  }
  else if ( (address >= 0x38) && (address <= 0x3F))
  {
    return ADDR_ACC_DCC;
  }
  else if ( (address >= 0x40) && (address <= 0x70))
  {
    return ADDR_MFX;
  }
  else if ( (address >= 0xC0) && (address <= 0xFF))
  {
    return ADDR_DCC;
  }
  else
  {
    return 0;
  }
}

//DCC専用
boolean DSCoreLib::WriteConfig_Ops(word address, word number, uint8_t value)
{
  uint8_t aPackets[5] = {0, 0, 0, 0, 0};
  uint8_t aLen;
  uint8_t aType;
  uint8_t aDecodedAddr1;
  uint8_t aDecodedAddr2;
  word aAddress;
  uint8_t aCVValue;
  uint8_t aPowerOnEnd;

  /* Address decode */

  Main_decodeAddress(address, &aType, &aDecodedAddr1, &aDecodedAddr2, &aLen);

  /* CV書き込み */
  aAddress = number - 1; /* 1-1024を、0-1023に変換 */
  aCVValue = value;

  /* CVパケット生成(書き込み指令) */
  if ( aLen == 2)
  {
    aPackets[0] = aDecodedAddr1;
    aPackets[1] = aDecodedAddr2;

    // 1110CCVV 0 VVVVVVVV 0 DDDDDDDD

    aPackets[2] = 0b11100000 | 0b00001100 | ((aAddress >> 8) & 0b11);
    aPackets[3] = aAddress & 0b11111111;
    aPackets[4] = aCVValue;
    aLen = 5;
  }
  else
  {
    aPackets[0] = aDecodedAddr1;
    aPackets[1] = 0b11100000 | 0b00001100 | ((aAddress >> 8) & 0b11);
    aPackets[2] = aAddress & 0b11111111;
    aPackets[3] = aCVValue;
    aLen = 4;
  }

    sendPackets(aPackets, aLen, DCCPREAMBLE_DEFAULT);

  return true;
}



/* データ取得後の処理 */
boolean DSCoreLib::CVReadPostProcess(uint8_t *value)
{
	uint8_t aDetectedCnt[16];
	long aData = 0;
	uint8_t aData_Count = 0;
	boolean aRet = false;
		
	for( uint8_t k = 0; k < 16; k++)
	{
		
		/* 配列に保存する */
		aDetectedCnt[aData_Count] = cvval_max[aData_Count];
		
		//Serial.print(cvval_max[aData_Count]);
		//Serial.println(",");
		
		//検出できたビットを立てていく(合計16回)
		if( cvval_max[aData_Count] >= gThresholdCurrent)
		{
			
			//データを立てる
			aData = aData + (1 << aData_Count);
		}
		
		aData_Count++;
		
	}

	/* upper and lower byte calculation */
	uint8_t aData_high = (aData >> 8 ) & 0xFF;
	uint8_t aData_low = aData & 0xFF;
	uint8_t aData_low_not = (~(uint8_t)aData_low) & 0xFF;

	/* データチェック */
#ifdef DEBUG
	Serial.println("");
	
	Serial.print(F("Data(BIN)="));
	Serial.println(aData & 0xFFFF, BIN);

		
	Serial.print(F("AddressCheck: 0bit="));
	Serial.print(aData_low_not, DEC);
	Serial.print(F(", 1bit="));
	Serial.println(aData_high, DEC);
#endif
	
	
	if( aData_high == aData_low_not)
	{
#ifdef DEBUG
		Serial.println(F("CV read Successed!"));
#endif
		
		*value = aData_low_not;
		
		aRet = true;
	}
	else
	{
		//リベンジ
		if( GetCVValueFromMap(aDetectedCnt, value, &aData_high, &aData_low) == true)
		{
#ifdef DEBUG
			Serial.println(F("CV read recovery Successed!"));
#endif
		
			aRet = true;
		}
		else
		{
			//Nothing to do
		}
		
		
		if( aRet == false)
		{
#ifdef DEBUG
			Serial.println(F("CV read Failed!"));
#endif
			*value = 0;
			
			aRet = false;
		}
	}

	
	return aRet;

}



bool DSCoreLib::GetCVValueFromMap(uint8_t *inCntData, uint8_t *outpValue, uint8_t *outpHighByte, uint8_t *outpLowByte)
{
	
	word aSum = 0;
	word aMax = 0;
	word aMin = 9999;
  word aCenter = 0;
	word aData = 0;
	word aThreshold = 0;
	uint8_t i, j;
  uint8_t aTemp = 0;
  uint8_t aDataTemp[SIZE_CVBITS];
	
#ifdef DEBUG
	Serial.print(F("Bits->"));
#endif

  //バブルソート(そのうち違うソートにする)

  for (i = 0; i < SIZE_CVBITS - 1; i++)
  {
    aDataTemp[i] = inCntData[i];
  }

  for (i = 0; i < SIZE_CVBITS - 1; i++)
  {
    for (j = SIZE_CVBITS - 1; j >= i + 1; j--)
    {
      if (aDataTemp[j] < aDataTemp[j - 1])
      {
         
         aTemp = aDataTemp[j];
         aDataTemp[j] = aDataTemp[j - 1];
         aDataTemp[j - 1] = aTemp;
      
      }
    }
  }

  //最小値、最大値、中央値を求める
  aMax = aDataTemp[0];
  aMin = aDataTemp[SIZE_CVBITS - 1];
  aCenter = (aDataTemp[7] + aDataTemp[8])>> 1; //７個目と８個目を足して二で割る

  // 0, 1, 2, 3, 4, 5, 6, 7,
  // 8, 9,10,11,12,13,14,15,
	
	for( i = 0; i < SIZE_CVBITS; i++)
	{
		aSum = aSum + (word)inCntData[i];
		
		if( inCntData[i] > aMax)
		{
			aMax = inCntData[i];
		}
		
		if( inCntData[i] < aMin)
		{
			aMin = inCntData[i];
		}
		
#ifdef DEBUG
		Serial.print(inCntData[i]);
		Serial.print(",");
#endif
		
	}
	
#ifdef DEBUG
	Serial.println("");
#endif
	
	aThreshold = aCenter;//中央値にする
	
	for( i = 0; i < SIZE_CVBITS; i++)
	{
		if( inCntData[i] < aThreshold)
		{
			//データを立てる
			aData = aData + (1 << i);
		}
	}
	
	uint8_t aData_high = (aData >> 8 ) & 0xFF;
	uint8_t aData_low = aData & 0xFF;
	uint8_t aData_low_not = (~(uint8_t)aData_low) & 0xFF;
	
#ifdef DEBUG
	Serial.print("DataB(BIN)=");
	Serial.println(aData & 0xFFFF, BIN);
	//Serial.print("Recovered Value =");
	//Serial.println(aData_low);
#endif
	
	if( aData_high == aData_low_not)
	{
		*outpValue = aData_low;
		return true;
	}
	else
	{
		return false;
	}

}

word DSCoreLib::GetCurrent(word inAvg)
{
	
	
	//ADCの選択(電流検出を選択しておく)
	adc_select_input(PIN_CURRENT_A2);

	//ADC読み出し
	long aCurrent = adc_read();
	
	
	if( aCurrent > inAvg)
	{
		aCurrent = aCurrent - inAvg;
	}
	else
	{
		aCurrent = 0;
	}
	
	return aCurrent;
}
	
word DSCoreLib::GetCurrentAvg()
{
	long aCurrent = 0;
	
	//ADCの選択(電流検出を選択しておく)
	adc_select_input(PIN_CURRENT_A2);
	
	for( uint8_t i = 0; i < 16; i++)
	{
		//ADC 12bit読み出し
		aCurrent = aCurrent + (long)adc_read();
	}
	
	return aCurrent >> 4;
}

/**
 * @brief 送信データの残りの長さを取得
 * 
 * @return word 送信データの残りの長さ
 */
word DSCoreLib::IsScanDataLen()
{
	word aRet = 0;
	
	if( gTxBufNum >= index_send)
	{
		aRet = gTxBufNum - index_send;
	}
	else
	{
		//Nothing to do
	}
	
	
	return aRet;
}

/**
 * @brief 送信データしたデータ長を取得
 * 
 * @return word 送信データの長さ
 */
word DSCoreLib::IsScanDataSent()
{

	return index_send;
}



/**
 * @brief 送信回数チェック
 * 
 * @return true 二回送信である
 * @return false 二回送信ではない
 */
bool DSCoreLib::CheckTwiceSend()
{
	//二回送信モードかどうかをチェック
	if( (gTxBufNum >= index_send) && (flag_send_twice > 0) && (len_send_twice > 0))
	{
		flag_send_twice--;
		
		return true;
	}
	else
	{
		return false;
	}
}

/**
 * @brief 送信データを4バイト(32bit)取得
 * 
 * @return uint32_t 送信データ(32bit)
 */
uint32_t DSCoreLib::ScanData()
{
	uint32_t aData = 0;
	
	if( gTxBufNum >= index_send)
	{
		aData = gTxBuf[index_send];
		
		index_send++;
	}
	
	//Serial.print(gTxBufNum);

	return aData;
}

/**
 * @brief 送信データをクリア
 * 
 */
void DSCoreLib::ClearPacketStack()
{
	stacked_bit_num = 0;
	gTxBufNum = 0;
	index_send = 0;
	len_send_twice = 0;
	
	for( int i = 0; i < SIZE_TXBUF; i++)
	{
		gTxBuf[i] = 0x00000000;
	}
	
}

/**
 * @brief 送信データを1埋め
 * 
 */
void DSCoreLib::FillPacketBit()
{
	word remainBit = 32 - (stacked_bit_num & 0b11111);
	
	if( remainBit > 0 )
	{
		for( int i = 0; i < remainBit; i++)
		{
			gTxBuf[gTxBufNum] = gTxBuf[gTxBufNum] + (1 << (stacked_bit_num - (gTxBufNum << 5)));
			stacked_bit_num++;
		}
	}
	
}

/**
 * @brief スタック整理・送信位置調整
 * 
 */
void DSCoreLib::PreparePacketStack()
{

	for( int i = 0; i < SIZE_TXBUF; i++)
	{
		gTxBuf[i] = 0x00000000;
	}
	
	//前回の最後のデータを最初に移動
	if(len_last_sendBits != 0)
	{
		gTxBuf[0] = data_end32_last_send;//gTxBuf[len_last_sendTx];
	}

/*
	Serial.print(" COPIED");
	Serial.print(len_last_sendTx);
	Serial.print(",");
	Serial.println(len_last_sendBits);
	*/

	//データ追加可能開始位置を調整
	stacked_bit_num = len_last_sendBits;
	gTxBufNum = 0;
	index_send = 0;
	len_send_twice = 0;
	
}

/**
 * @brief ビットデータを送信データにセット
 * 
 * @param aBit 送信データ(1bit)
 */
void DSCoreLib::pushPacketBit(uint8_t aBit)
{
	gTxBuf[gTxBufNum] = gTxBuf[gTxBufNum] + (aBit << (stacked_bit_num - (gTxBufNum << 5)));
	stacked_bit_num++;
	
	if( (stacked_bit_num >> 5) > gTxBufNum )
	{
		gTxBufNum++;
	}

}

/**
 * @brief バイトデータを送信データにセット
 * 
 * @param aByte 送信データ(1byte)
 */
void DSCoreLib::pushPacketByte(uint8_t aByte)
{
	for( int i = 0; i < 8; i++)
	{
		pushPacketBit((aByte >> (7 - i)) & 1);
	}	
}

/**
 * @brief パケット送信データ生成・セット
 * 
 * @param inPackets 送信データのバイト列
 * @param inLen 送信データの長さ(Byte単位)
 * @param inPreambleCnt プリアンブル数
 * @param inCutOut カットアウトの有無
 */
void DSCoreLib::sendPackets(uint8_t *inPackets, uint8_t inLen, uint8_t inPreambleCnt)
{
	int i;

	/* Calculate CRC */
	uint8_t aCRC = DS_CalculateCRC(inPackets, inLen);

	/* Send Preamble */
	for( i = 0; i < inPreambleCnt; i++ )
	{
		pushPacketBit(1);
	}
	
	for( i = 0; i < inLen; i++)
	{
		/* StartBit */
		pushPacketBit(0);
		
		/* Send data */
		pushPacketByte( inPackets[i]);
	}

	/* StartBit */		
	pushPacketBit(0);
	
	/* SendCRC */
	pushPacketByte(aCRC);
		
	/* EndBit */		
	pushPacketBit(1);
	
	len_send_twice = gTxBufNum;
	len_last_sendTx = gTxBufNum;
	len_last_sendBits = stacked_bit_num & 0b11111;
	data_end32_last_send = gTxBuf[gTxBufNum];

	/*
	//DEBUG
	Serial.print("SENT SIZE: ");
	Serial.print(len_last_sendTx);
	Serial.print(" / ");
	Serial.println(len_last_sendBits);
	*/
}

/**
 * @brief リセットパケットの送信
 * 
 * @param inPreambleCnt プリアンブル数
 */
void DSCoreLib::sendInitialPacket(uint8_t inPreambleCnt)
{
	uint8_t aPackets[5] = {0,0,0,0,0};
	
	/* DCC Resetパケット登録（初回のみ） */
	aPackets[0] = 0b00000000;
	aPackets[1] = 0b00000000;
	aPackets[2] = 0b00000000;
	
	sendPackets(aPackets, 2, inPreambleCnt);
	
}

/**
 * @brief CV書き込み(Direct Mode)
 * 
 * @param number CV番号[1-1024]
 * @param value  CV値
 * @return boolean 書き込みAck状態
 */
boolean DSCoreLib::WriteConfig_Dir( word number, uint8_t value)
{
	uint8_t aPackets[5] = {0,0,0,0,0};
	word aAddress;
	uint8_t aCVValue;
	bool aRet = false;
	word aTempVal;
	word aMaxVal;
	word aMinVal;
	uint32_t aBaseTime;
	word aCurrentOffset; //電流センサオフセット
	
	/* CV書き込み */
	aAddress = number;
	aCVValue = value;
	
	//オフセット調整用
	aCurrentOffset = GetCurrentAvg();
		
	//ピン設定を手動操作に切り替え
	InitPins();

	/* 線路に電源投入 */
	DS_Power(1);
		
	/*DCC*/
	aAddress = aAddress - 1;/* 1-1024を、0-1023に変換 */
	
	
	/* Power-on-cycle */
	for( uint8_t i = 0; i < 20; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}
	
	/* リセットパルスを3回以上送る */
	for( uint8_t i = 0; i < 6; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}
		
	/* CVパケット生成(書き込み指令) */
	aPackets[0] = 0b01110000 | 0b00001100 | ((aAddress >> 8) & 0b11);
	aPackets[1] = aAddress & 0b11111111;
	aPackets[2] = aCVValue;


	//初期化
	gCVAckMaxVal = 0;
	gCVAckMinVal = 0xFFFF;
	gCVAckAdcOffset = aCurrentOffset;

	//1ms周期割り込み開始
	add_repeating_timer_ms( -1, timer_callback_cvack_1ms, NULL, &timer_1ms_cvack_task ); // Timerの起動
	
	/* CV指令を5回以上送る */
	for( uint8_t i = 0; i < 10; i++)
	{
		DCC_SendPacketsEx(aPackets, 3, DCCPREAMBLE_CV);
		
		if( gCVAckMaxVal >= gThresholdCurrent)
		{
			aRet = true;
			break;
		}		
	}

	//CV Ack検出割り込みを終了
	cancel_repeating_timer(&timer_1ms_cvack_task);
		
	/* リセットパルスを6回以上送る */
	for( uint8_t i = 0; i < 32; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}



	/* 電源を切る */
	DS_Power(0);

	return aRet;
}


/**
 * @brief CV読み込み処理(DirectMode)
 * 
 * @param number CV番号[1-1024]
 * @param value 得られる値(ポインタ)
 * @return boolean 読み取り結果
 */
boolean DSCoreLib::ReadConfig(word number, uint8_t *value)
{
	boolean aRet = false;
	
	//ピン設定を手動操作に切り替え
	InitPins();
	
	/* 線路に電源投入 */
	DS_Power(1);

	/* Power-on-cycle */
	for( uint8_t i = 0; i < 20; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}
	
	/* リセットパルスを3回以上送る */
	for( uint8_t i = 0; i < 6; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}	

	/* 詳細関数を呼び出す */
	aRet = ReadConfigEx(number, value);
	
	/* いったん電源を切る */
	DS_Power(0);
	
	return aRet;
}

/**
 * @brief CV読み込み処理の実体(DirectMode)
 * 
 * @param number CV番号[1-1024]
 * @param value 得られる値(ポインタ)
 * @return boolean 読み取り結果
 */
boolean DSCoreLib::ReadConfigEx(word number, uint8_t *value)
{
	uint8_t aDetectedCnt[16];
	long aData = 0;
	uint8_t aData_Count = 0;
	word aCurrentOffset; //電流センサオフセット
	word aMaxVal;
	word aMinVal;
	boolean aRet = false;
	
	int aCntReset = 3;
	int aCntCV = 4;
	
	//オフセット調整用
	aCurrentOffset = GetCurrentAvg();

	for( uint8_t d = 0; d < 2; d++)
	{
		
		for( uint8_t k = 0; k < 8; k++)
		{
			
			getCVData(number, &aMaxVal, &aMinVal, d, k, aCntReset, aCntCV, aCurrentOffset);

			
			/* 配列に保存する */
			aDetectedCnt[aData_Count] = aMaxVal;
			
			//検出できたビットを立てていく(合計16回)
			if( aMaxVal >= gThresholdCurrent)
			{
				
				//データを立てる
				aData = aData + (1 << aData_Count);
			}
			
			aData_Count++;
			
		}
	}
	

	/* upper and lower byte calculation */
	uint8_t aData_high = (aData >> 8 ) & 0xFF;
	uint8_t aData_low = aData & 0xFF;
	uint8_t aData_low_not = (~(uint8_t)aData_low) & 0xFF;

	/* データチェック */
#ifdef DEBUG
	Serial.println("");
	
	Serial.print(F("Ioffset="));
	Serial.println(aCurrentOffset);
	Serial.print(F("Data(BIN)="));
	Serial.println(aData & 0xFFFF, BIN);

		
	Serial.print(F("AddressCheck: 0bit="));
	Serial.print(aData_low_not, DEC);
	Serial.print(F(", 1bit="));
	Serial.println(aData_high, DEC);
#endif
	
	
	if( aData_high == aData_low_not)
	{
#ifdef DEBUG
		Serial.println(F("CV read Successed!"));
#endif
		
		*value = aData_low_not;
		
		aRet = true;
	}
	else
	{
		//リベンジ
		if( GetCVValueFromMap(aDetectedCnt, value, &aData_high, &aData_low) == true)
		{
#ifdef DEBUG
			Serial.println(F("CV read recovery Successed!"));
#endif
			
			aRet = true;
		}
		else
		{
			
#ifdef DEBUG
			Serial.println(F("CV read retry"));
#endif
			
			for( uint8_t e = 0; e < gRetryCount; e++)
			{
				/* エラー箇所だけ読み取りし直し */
				
				aCntCV++;
				aCntReset++;
				
				for( uint8_t d = 0; d < 8; d++)
				{
					if( ((aData_high >> d) & 1) == ((aData_low >> d) & 1))
					{
						
						/* Bitが異なるものを取り直す */
						
						getCVData(number, &aMaxVal, &aMinVal, 0, d, aCntReset, aCntCV, aCurrentOffset);
						aDetectedCnt[d] = aMaxVal;
						
#ifdef DEBUG
						Serial.print(aMaxVal);
#endif
						delay(6);
						
						getCVData(number, &aMaxVal, &aMinVal, 1, d, aCntReset, aCntCV, aCurrentOffset);
						aDetectedCnt[d + 8] = aMaxVal;
						
#ifdef DEBUG
						Serial.print("=");
						Serial.print(aMaxVal);
						Serial.print(", ");
						Serial.println("");
#endif
						delay(6);
						
					}
				}
				
				/* 再度、読み直し */
				if( GetCVValueFromMap(aDetectedCnt, value, &aData_high, &aData_low) == true)
				{
#ifdef DEBUG
					Serial.println(F("CV read retry Successed!"));
#endif
					
					aRet = true;
					
					break;
				}
			}
			
			
			if( aRet == false)
			{
#ifdef DEBUG
				Serial.println(F("CV read Failed!"));
#endif
				*value = 0;
				
				aRet = false;
			}
		}
	}

	
	return aRet;

}

/**
 * @brief CVのbit読み出し処理
 * 
 * @param inAddress CV番号[1-1024]
 * @param inMaxData 得られた電流反応値の最大
 * @param inMinData 得られた電流反応値の最小
 * @param inD チェックするビット値
 * @param inK bit番号
 * @param inCntReset リセット出力回数
 * @param inCntCV CV読み出しパケット出力回数
 * @param inAdcOffset ADCのオフセット値
 */
void DSCoreLib::getCVData(word inAddress, word *inMaxData, word *inMinData, uint8_t inD, uint8_t inK, uint8_t inCntReset, uint8_t inCntCV, word inAdcOffset)
{
	
	uint8_t aPackets[5] = {0,0,0,0,0};
	word aAddress;
	uint8_t aThreshlodCount = 0;
	uint8_t aDetected = 0;
	
	/*DCC*/
	aAddress = inAddress - 1;/* 1-1024を、0-1023に変換 */
	
	/* CVパケット生成(書き込み指令) */
	aPackets[0] = 0b01110000 | 0b00001000 | ((aAddress >> 8) & 0b11);
	aPackets[1] = aAddress & 0b11111111;

	/* リセットパルスを3回以上送る */
	for( uint8_t i = 0; i < inCntReset; i++)
	{
		DCC_SendInitialPacket(DCCPREAMBLE_CV);
	}
	
	/* CVパケット生成(書き込み指令) */
	aPackets[2] = 0b11100000 | 0b00000000 | (inD == 0 ? 0 : 0b1000)  | inK;// 111KDBBB: K=0(BitVerify),D= 0 or 1, BBB=bit番号 
	
	//初期化
	gCVAckMaxVal = 0;
	gCVAckMinVal = 0xFFFF;
	gCVAckAdcOffset = inAdcOffset;
	
	//1ms周期割り込み開始
	add_repeating_timer_ms( -1, timer_callback_cvack_1ms, NULL, &timer_1ms_cvack_task ); // Timerの起動
	
	for( uint8_t i = 0; i < inCntCV; i++)
	{
		//Packet send (1回目～)
		DCC_SendPacketsEx(aPackets, 3, DCCPREAMBLE_CV);
		
		if( gCVAckMaxVal >= gThresholdCurrent)
		{
			//1回目 Ack応答のレスポンス
			DCC_SendInitialPacket(DCCPREAMBLE_CV);
			//2回目 Ack応答のレスポンス
			DCC_SendInitialPacket(DCCPREAMBLE_CV);
			
			break;
		}
		
	}
		
	//CV Ack検出割り込みを終了
	cancel_repeating_timer(&timer_1ms_cvack_task);
	
	*inMaxData = gCVAckMaxVal;
	*inMinData = gCVAckMinVal;
	
	
}
