/*********************************************************************
   Desktop Station air Sketch for RP2040 Hardware Platform

   Copyright (C) 2023 DesktopStation Co.,Ltd. / Yaasan
   https://desktopstation.net/
*/

// include the library code:
#include <SPI.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include "OzOLED.h"
#include "hardware/adc.h"
#include "hardware/flash.h"
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM.h"
#include "Functions.h"
#include "webapp.h"
#include "UIManager.h"
#include "ShareMemManager.h"
#include "ParamManager.h"
#include "dccpulsegen.h"
#include "board_pins.h"
#include "DSVersion.h"


//#define DEBUG
#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3
#define LEDSTATE_ANALOG 4

#define ENABLE_RAILCOM 1	//0: Disable, 1: Enable
#define THRESHOLD_LV_DSA 1100	// 4096d *(1.2 / 11.2) *9V / 3.3V
#define THRESHOLD_OV_DSA 2700	// 4096d *(1.2 / 11.2) *20V / 3.3V
#define TIMEOUT_SERIAL 30

#define PIODCC_STATE_IDLE		0
#define PIODCC_STATE_OUTDCC		1
#define PIODCC_STATE_OUTANA		2
#define PIODCC_STATE_OUTCV		3
#define PIODCC_STATE_ERROR		4

#define PIODCC_GEN_READY		0
#define PIODCC_GEN_DCCSEND		1
#define PIODCC_GEN_CVSEND		2


typedef struct
{
	uint8_t mOV: 1;
	uint8_t mOC: 1;
	uint8_t mLV: 1;
	uint8_t mUnk1: 1;
	uint8_t mUnk2: 1;
	uint8_t reserved: 3;
} DATA_STATUS;


/*WiFi Server */
char ssid[27];	// your network SSID (name)
char pass[] = "12345678";	// your network password (use for WPA, or use as key for WEP)
int keyIndex = 0;	// your network key index number (needed only for WEP)
int status = WL_IDLE_STATUS;

/*WiFi Server */
WiFiServer server(80);
WiFiClient clients[4];
ShareMemManager sharemem;
ParamManager paramman;

uint8_t buffer_sharedMem[SHAREMEM_SIZE + 1];
uint8_t buffer_lastcmd[SHAREMEM_SIZE];
uint8_t *buffer_sharedStatusMem;
uint8_t index_client = 0;

/* UI制御 */
UIManager UI;

/* OLED */
OzOLED LCD;



/*Decralation classes */
DSCoreLib DSCore;
uint8_t state_piodcc = PIODCC_STATE_IDLE;
uint8_t state_dccgen = PIODCC_GEN_READY;


/*LED functions */
uint8_t gLED_State = LEDSTATE_STOP;
uint8_t gLED_Toggle = 0;
uint8_t gLED_Counter = 0;

/*Sensor */
short gOffset_Current = 0;
word gEdc_data = 0;
short gCur_data = 0;
uint8_t gSwSensor = 0;
uint8_t gMaxS88Num = 0;

/*Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;
uint8_t gLastErrorNo = 0;
uint8_t gAliveSeq = 0;

/*Parser */
String gRequest;
String function;
word arguments[4];
uint8_t numOfArguments;
word gLocAddresses[4] = { 
	0,
	0,
	0,
	0
};

uint8_t numOfLocAddr = 0;
boolean result;

DATA_STATUS gStatus;

//Programmable IO (PIO)
PIO gPio;
uint gPio_SM;
uint gPio_PrgAddr;

//コア間データ受け渡し(WiFiからのコマンド)
String gRequest_wifi = "";

//Bidi couneter
uint8_t state_bidicutout = 0;
word gCounter_BiDi = 0;
word gInterval_BiDi = 0;
uint8_t gFlag_PacketType = 0;


//Task Schedule
unsigned long gPreviousTask100ms = 0;	//   50ms interval
unsigned long gPreviousTask500ms = 0;	//   50ms interval

//状態遷移
uint8_t state_500ms_interval = 0;

//task
repeating_timer_t timer_bidi29us_task; // timer変数の定義
repeating_timer_t timer_buttoninput_task; // timer変数の定義
repeating_timer_t timer_100ms_task; // timer変数の定義
repeating_timer_t timer_500ms_task; // timer変数の定義

//定周期処理
bool timer_callback_bidi29us(repeating_timer_t *rt);
bool timer_callback_buttoninput(repeating_timer_t *rt);
bool timer_callback_100ms(repeating_timer_t *rt);
bool timer_callback_500ms(repeating_timer_t *rt);

void ProcessParse();
void printString(const char *s, int x, int y);
void ReplyPowerPacket(uint8_t inPower);

void PIODCC_Init();
void PIODCC_StopPIO();
void PIODCC_StartPIO();
void PIODCC_ChangeProcess(uint8_t inNextState);
void dccrecvIrqHandler();



/***********************************************************************

   Boot setup

 ************************************************************************/

void setup()
{
	/*Debug message for PC */
	Serial.begin(115200);

	
	//初期化
	/*共有メモリクリア */
	sharemem.ClearShareMem();
	sharemem.LOCMNG_Clear();
	
	buffer_sharedStatusMem = sharemem.GetShareMemPointer();

	//アラーとピン
	gpio_init(PIN_ALERT);
	gpio_set_dir(PIN_ALERT, GPIO_IN);
	gpio_disable_pulls(PIN_ALERT);

	//analog
	analogReadResolution(12);
	
	/* UI Configuration */
	UI.LCD = LCD;
	UI.Init();
	UI.callback_dscommand = callback_control;
	
	/*DSCore初期化 */
	DSCore.Init();
	DSCore.gCutOut = ENABLE_RAILCOM;
	
	//PIO DCC初期化
	PIODCC_Init();

	/*線路に電源遮断 */
	DS_Power(0);

	/*Dammy sleep */
	CalcCurrentOffset();

	//Hardware seriesのチェック
	gThreshold_OV = THRESHOLD_OV_DSA;
	gThreshold_LV = THRESHOLD_LV_DSA;


	/*LED turn on */
	changeLED();

	//SSID
	uint8_t aUniqueID[8];
	flash_get_unique_id(aUniqueID);
	sprintf(ssid, "dsairlite_%0x%0x%0x%0x%0x%0x%0x%0x", aUniqueID[0],aUniqueID[1],aUniqueID[2],aUniqueID[3],aUniqueID[4],aUniqueID[5],aUniqueID[6],aUniqueID[7]);

	UI.RegisterWifi(ssid, pass);
	
	//WiFi
	setup_wifi();    // 20231229 HMX : setup2から
	setup_server();  // 20231229 HMX : setup2から

	Serial.println(F("100 Ready"));

	//Reset task
	gPreviousTask100ms = millis();
	gPreviousTask500ms = millis();

	//設定ロード
	paramman.Clear();


}

boolean parse()
{
	int aPosSlash = -1;

	char aTemp[SHAREMEM_SIZE];
	gRequest.toCharArray(aTemp, SHAREMEM_SIZE);

	int lpar = gRequest.indexOf('(');
	if (lpar == -1)
	{
		return false;
	}

	function = String(gRequest.substring(0, lpar));

	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = gRequest.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;

	while (comma != -1)
	{
		String tmp = gRequest.substring(offset, comma);
		tmp.trim();

		/*スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);

		if (aPosSlash > 0)
		{
			tmp = tmp + "/";

			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;

			//スラッシュはアドレス部だけに許可
			while (aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();

				if (numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}

				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}

			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
	}

	int rpar = gRequest.indexOf(')', offset);
	
	while (rpar == -1)
	{
		return false;
	}

	if (rpar > offset)
	{
		String tmp = gRequest.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch(uint8_t inCommandMode)
{
	uint8_t aValue;
	boolean aResult;
	uint8_t aDir = 0;
	
	if (function == "who")
	{
		
		Serial.print("PID");
		Serial.print(",");
		Serial.print(DECODER_PRODUCTID, DEC);
		Serial.println(";");
		
		Serial.print("FVER");
		Serial.print(",");
		Serial.print(FIRMWARE_VERSION, DEC);
		Serial.println(";");
		
		Serial.print("FIN");
		Serial.println(";");
		
		return true;
	}
	else if ((function == "setLocoDirection") || ((function == "DI")))
	{
		if (arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;

			//異常値はREV固定
			if (aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}

		if (numOfLocAddr > 0)
		{
			//重連対応
			for (int i = 0; i < numOfLocAddr; i++)
			{
				sharemem.LOCMNG_SetLocoDir(gLocAddresses[i], aDir);

				if( inCommandMode != 1)
				{
					UI.NotifyDCCDirection(gLocAddresses[i], aDir);
				}
				
				aResult = DSCore.SetLocoDirection(gLocAddresses[i], aDir);
			}
			
			aResult = true;
		}
		else
		{
			//非重連
			sharemem.LOCMNG_SetLocoDir(arguments[0], aDir);

			if( inCommandMode != 1)
			{
				UI.NotifyDCCDirection(arguments[0], aDir);
			}			
			
			aResult = DSCore.SetLocoDirection(arguments[0], aDir);
			
		}

		return aResult;
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		sharemem.LOCMNG_SetLocoFunc(arguments[0], arguments[1], arguments[2]);
		
		if( inCommandMode != 1)
		{
			UI.NotifyDCCFunction(arguments[0], arguments[1], arguments[2]);
		}
		
		
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		//テーブル更新
		//MM2: 1(0x3000) - 320(0x3140)
		//DCC: 1(0x3800) - 2044(0x3FFC)

		word aAccAddr = 0;

		if ((arguments[0] >= 0x3000) && (arguments[0] <= 0x3140))
		{
			aAccAddr = arguments[0] - 0x3000;
		}
		else if ((arguments[0] >= 0x3800) && (arguments[0] <= 0x3FFC))
		{
			aAccAddr = arguments[0] - 0x3800;
		}

		if ((aAccAddr >> 3) <= 31)
		{
			sharemem.LOCMNG_UpdateAccBuffer(aAccAddr, arguments[1]);
			sharemem.LOCMNG_UpdateAcc(aAccAddr >> 3);
		}

		//UIへの反映
		if( inCommandMode != 1)
		{
			UI.NotifyAccessory(aAccAddr, arguments[1]);
		}		

		//ポイントパケット出力登録

		return DSCore.SetTurnout(arguments[0], (uint8_t) arguments[1]);
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		if (arguments[0] == 1)
		{
			//アナログに切り替わっている場合に再設定
			DSCore.InitPins();
			
			//PIODCC開始
			PIODCC_ChangeProcess(PIODCC_STATE_OUTDCC);
			
			//電流オフセット計測
			CalcCurrentOffset();

			/*状態変更 */
			gLED_State = LEDSTATE_RUN;

			//FlashAirのバッファにセット
			sharemem.RegisterShareMem(SMEM_POWER, 1);
			sharemem.RegisterShareMem(SMEM_ERROR, 1);
		}
		else
		{ /*状態変更 */
			gLED_State = LEDSTATE_STOP;
			sharemem.RegisterShareMem(SMEM_POWER, 0);
			
			//PIODCC停止
			PIODCC_ChangeProcess(PIODCC_STATE_IDLE);
		}

		/*LED状態の変更 */
		gLED_Counter = 0;
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		if (numOfLocAddr > 0)
		{
			//重連対応
			for (int i = 0; i < numOfLocAddr; i++)
			{
				sharemem.LOCMNG_SetLocoSpeed(gLocAddresses[i], arguments[1]);
				aResult = DSCore.SetLocoSpeedEx(gLocAddresses[i], arguments[1], arguments[2]);
			}

			aResult = true;
		}
		else
		{
			//非重連
			sharemem.LOCMNG_SetLocoSpeed(arguments[0], arguments[1]);
			aResult = DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
			
		}

		return aResult;
	}
	else if (function == "DC")
	{
		//PRM0=Speed(0-1023), PRM1=Dir(0,1=FWD, 2=REV)
		
		//32kHz出力設定
		//analogWriteFreq(32000);
		
		if( arguments[0] > 0)
		{
		
			SetAnalogSpeed(arguments[0], arguments[1]);
		}
		else
		{
			SetAnalogSpeed(0, 0);
		}
		
		return true;
	}
	else if ((function == "reset") || ((function == "RS")))
	{
		#ifdef DEBUG
		Serial.println(F("100 Ready"));
		#endif

		/*リセットパケットを投げる */
		if (gLED_State == LEDSTATE_RUN)
		{
			//内部のデータバッファを削除する
			DSCore.Clear();

			//リセット
			DSCore.SendReset();
		}

		return true;
	}

	/*reset */
	else if ((function == "setPing") || ((function == "PG")))
	{
		/*Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println(F("@DSG,00,03,00,02,00,00,00,00"));

		/*DSair2互換 */
		Serial.print("@DSS,");

		for (word i = 0; i <= 264; i++)
		{
			Serial.print(char(buffer_sharedStatusMem[i]));
		}

		Serial.println("");

		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		

		//Clear last CV data
		sharemem.WriteSharedMem_CVReply(0, 0);
	
		//CV Read（Software Pulse Gen)
		aValue = 0;
		aResult = DSCore.ReadConfig(arguments[1], &aValue);

		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");

		if (aResult == true)
		{
			Serial.print(aValue);
			UI.SetResultCVValue(aValue);
			
			//Write to FlashAir
			sharemem.WriteSharedMem_CVReply(arguments[1], aValue);
		}
		else
		{
			Serial.print(0xFFFF);
			UI.SetResultCVValue(0xFFFF);
			
			//Write to FlashAir
			sharemem.WriteSharedMem_CVReply(0, 255);
		}

		Serial.println(",");
		
		
		//出力停止・応答は処理完了後に他のタスクで対応。
		
		return true;
	}
	else if (function == "gS8")
	{
		return true;
	} /*getS88 for Wi-Fi */
	else if (function == "getS88")
	{
		//Send a S88 sensor reply
		Serial.print("@S88,");

		Serial.println("0000");

		return true;
	} /*getS88 */
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");

		//Clear last CV data
		sharemem.WriteSharedMem_CVReply(0, 0);
		
		//CV Write（Software Pulse Gen)
		DSCore.WriteConfig_Dir(arguments[1], arguments[2]);
		
		//Write to FlashAir
		sharemem.WriteSharedMem_CVReply(arguments[1], arguments[2]);

		return true;
	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		/*Railcom CutOut */
		DSCore.gCutOut = (arguments[0] == 1) ? 1 : 0;

		return true;
	}
	else
	{
		return false;
	}
}

/***********************************************************************

   Timer Interval

 ************************************************************************/

void loop()
{

	//WiFi処理
	loop_wifi();

	if ((millis() - gPreviousTask100ms) >= 100)
	{
		//Reset task
		gPreviousTask100ms = millis();
		
		//UI関連
		UI.Process();

	}

	if ((millis() - gPreviousTask500ms) >= 500)
	{
		//Reset task
		gPreviousTask500ms = millis();


		switch(state_500ms_interval)
		{
			case 0:
				state_500ms_interval++;
				break;
			case 1:
				//割り込み起動
				add_repeating_timer_ms( -50, timer_callback_100ms, NULL, &timer_100ms_task ); // Timerの起動
				add_repeating_timer_ms( -500, timer_callback_500ms, NULL, &timer_500ms_task ); // Timerの起動

				state_500ms_interval++;
				break;
			case 2:				
				//LED Control
				taskLED();

				//送信回数
				sharemem.RegisterShareMem(SMEM_SEQ, gAliveSeq);

				if (gAliveSeq >= 99)
				{
					gAliveSeq = 0;
				}
				else
				{
					gAliveSeq++;
				}
				break;
		}


	}




	int aReceived = receiveRequest();

	if (aReceived > 0)
	{
		Serial.println(gRequest);
		
		ProcessParse();
		/*Reply to Desktop Station */
	}
	else
	{
		/*Nothing to do */

	}

}

void ProcessSensors()
{
	long aEdc;
	signed long aCur;
	uint8_t aErrNo = 1;
	
	switch(gLED_State)
	{
		
	case LEDSTATE_ERR:
		
		break;
		
	case LEDSTATE_STOP:
		
		break;
		
	case LEDSTATE_RUN:
		
		if (gSwSensor == 0)
		{
			gSwSensor = 1;
			gEdc_data = UI.GetVdc();

			/*OV check */
			if (gEdc_data > gThreshold_OV)
			{
				Serial.print("ERR: OV,");
				Serial.println(gEdc_data);
				aErrNo = 4;
			}

			/*LV check */
			if (gEdc_data < gThreshold_LV)
			{
				Serial.print("ERR: LV,");
				Serial.println(gEdc_data);
				aErrNo = 2;
			}

			// 3.3*(11.2/1.2) *X / 4096
			aEdc = (((long) gEdc_data * 308) >> 12);

			if (aEdc > 255)
			{
				aEdc = 255;
			}

			//Sensor Write
			sharemem.RegisterShareMem(SMEM_EDC, aEdc);
		}
		else
		{
			gSwSensor = 0;

			//ADC読み出し
			gCur_data = UI.GetCurrent();
			
			//
			aCur = ((long) gCur_data * 702) / 4095;
			aCur /= 10;

			if (aCur > 125)
			{
				aCur = 125;
			}

			sharemem.RegisterShareMem(SMEM_CUR, aCur);
		}

		/*DSairR2 HW  */
		if (gpio_get(PIN_ALERT) == 1)
		{
			//Over current at TB6642FG
			aErrNo = 8;
		}

		/*エラー処理 */
		if (aErrNo > 1)
		{
			if (DSCore.IsPower() == 1)
			{
				PowerOffByErr(aErrNo);
				//Serial.println(aErrNo);
				
				//UIに通知
				UI.NotifyError(aErrNo);
				
				
			}

			//FlashAirのバッファにセット
			sharemem.RegisterShareMem(SMEM_ERROR, aErrNo);
		}
		break;
	}
}

void ProcessParse()
{
	
	if (parse())
	{
		#ifdef DEBUG
		Serial.println(gRequest);
		#endif

		if (dispatch(0))
		{
			ApplyCommandUI();
			
			Reply200();
		}
		else
		{
			Reply300();
		}
	}
	else
	{
		Reply301();
	}
}

void Reply200(void)
{
	Serial.println(F("200 Ok"));
}

void Reply300(void)
{
	Serial.println(F("300 Cmd err"));
}

void Reply301(void)
{
	Serial.println(F("301 Stx err"));
}

void Reply302(void)
{
	Serial.println(F("302 Timeout"));
}

void Reply303(void)
{
	Serial.println(F("303 Unkn err"));
}

void Reply304(void)
{
	Serial.println(F("304 Ser err"));
}

void PowerOffByErr(uint8_t inErrNo)
{
	//PIODCC停止
	PIODCC_ChangeProcess(PIODCC_STATE_IDLE);
	
	DSCore.SetPower(0);
	ReplyPowerPacket(0);

	/*状態変更 */
	gLED_State = LEDSTATE_ERR;
	gLED_Counter = 0;
	gLastErrorNo = inErrNo;

	/*FlashAirにも反映 */
	sharemem.RegisterShareMem(SMEM_POWER, 0);

}

void ReplyPowerPacket(uint8_t inPower)
{
#ifdef DEBUG
	/*Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
#endif
}

void changeLED(void)
{
	switch (gLED_State)
	{
		case LEDSTATE_STOP:
			//gpio_put(PIN_RUNLED, 1);
			break;

		case LEDSTATE_RUN:
			//gpio_put(PIN_RUNLED, 0);
			break;

		case LEDSTATE_ANALOG:
			break;

		case LEDSTATE_ERR:
			break;

		case LEDSTATE_NOSD:
			break;
	}
}

void taskLED(void)
{
	/*カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter &0b1111;

	//LED操作
	switch (gLED_State)
	{
		case LEDSTATE_STOP:

			if (gLED_Counter <= 3)
			{
				//gpio_put(PIN_RUNLED, 1);
			}
			else
			{
				//gpio_put(PIN_RUNLED, 0);
			}

			break;

		case LEDSTATE_RUN:
			break;

		case LEDSTATE_ERR:

			if (gLED_Toggle == 0)
			{
				//gpio_put(PIN_RUNLED, 0);
			}
			else
			{
				//gpio_put(PIN_RUNLED, 1);
			}

			gLED_Toggle = (~gLED_Toggle) &1;
			break;

		case LEDSTATE_ANALOG:
			if ((gLED_Counter == 0) || (gLED_Counter == 8))
			{
				//gpio_put(PIN_RUNLED, 0);
			}
			else
			{
				//gpio_put(PIN_RUNLED, 1);
			}

			break;

		case LEDSTATE_NOSD:

			if (gLED_Counter <= 0)
			{
				//gpio_put(PIN_RUNLED, 1);
			}
			else if (gLED_Counter == 2)
			{
				//gpio_put(PIN_RUNLED, 1);
			}
			else
			{
				//gpio_put(PIN_RUNLED, 0);
			}

			break;
	}
}

//FFFF,1,FFFFFFFF

void CalcCurrentOffset()
{
	unsigned short aCurrentSum = 0;

	if (gOffset_Current == 0)
	{
		/*直流中間に電圧が印加されてないと、実際の動作時のADCのオフセット誤差が異なる */
		if (analogRead(PIN_EDC) >= THRESHOLD_LV_DSA)
		{
			gOffset_Current = DSCore.GetCurrentAvg();
		}		
	}

	/*検出できない場合は512をオフセット値にセット */
	if (gOffset_Current == 0)
	{
		gOffset_Current = 0;
	}
}


int receiveRequest()
{
	char buffer[64];
	int i = 0;
	char aByte;
	int aResult = -1;

	unsigned long time = millis();

	/*Check serial buffer */
	if (Serial.available() > 0)
	{
		//From Serial comm
		while (1)
		{
			if (Serial.available() > 0)
			{
				aByte = Serial.read();

				// Write to text buf.
				buffer[i] = aByte;

				// check last identification of text end.
				if (aByte == '\n')
				{
					aResult = i;
					break;
				}

				// increment
				i = i + 1;

				// buf over check
				if (i > 64)
				{
					break;
				}
			}
			else
			{ 	/*Check timeout. */
				if (millis() > time + 300)
				{
					aResult = -1;
					break;
				}
			}
		}
		
		//バッファをリクエストにセット
		if (aResult > 0)
		{
			buffer[i] = '\0';

			gRequest = String(buffer);
		}
		else
		{
			//Nothing to do
		}
		
	}
	else
	{
		
		String aRequestFromWifi = gRequest_wifi;
		
		if( aRequestFromWifi.length() > 0)
		{
			//From Wifi
			//Core 1と共用変数の処理
			aResult = aRequestFromWifi.length();
			gRequest = aRequestFromWifi;
			gRequest_wifi = "";
			
		}
		else
		{
			/*Not received. */
			aResult = 0;
			gRequest = "";
		}
		
		
	}



	return aResult;
}

void changedTargetSpeed(const word inAddr, const word inSpeed, const bool inUpdate)
{
	uint8_t aProtocol = 2;

	if (inAddr <= 255)
	{
		aProtocol = 0;
	}

	if (inUpdate == true)
	{
		DSCore.SetLocoSpeedEx(inAddr, inSpeed, aProtocol);
	}
}

void changedTargetFunc(const word inAddr, const uint8_t inFuncNo, const uint8_t inFuncPower, const bool inUpdate)
{
	if (inUpdate == true)
	{
		DSCore.SetLocoFunction(inAddr, inFuncNo, inFuncPower);
	}
}

void changedTargetDir(const word inAddr, const uint8_t inDir, const bool inUpdate)
{
	if (inUpdate == true)
	{
		DSCore.SetLocoDirection(inAddr, inDir);
	}
}

void changedTurnout(const word inAddr, const uint8_t inDir, const bool inUpdate)
{
	if (inUpdate == true)
	{
		DSCore.SetTurnout(inAddr, inDir);
	}
}

void changedPowerStatus(const uint8_t inPower)
{
	DSCore.SetPower(inPower);

}

//---------------------------------------------------------------------------
// PIO関連
//---------------------------------------------------------------------------

void PIODCC_Init()
{
	//PIOのch 0を確保
	gPio = pio0;

	//アセンブラのPIOプログラムへのアドレス
	gPio_PrgAddr = pio_add_program(gPio, &dcc_pulse_gen_program);

	// 空いているステートマシンの番号を取得
	gPio_SM = pio_claim_unused_sm(gPio, true);
	
}


void PIODCC_StopPIO()
{
	
	//割り込み不許可
	irq_set_enabled(PIO0_IRQ_0, false);
	
    pio_sm_set_enabled(gPio, gPio_SM, false);
	
	//ピン出力を変更し、出力OFF
	DSCore.InitPins();
	
	
	//Serial.println("STOP PIO");

}

void PIODCC_StartPIO()
{
	
	//DCC pulse input
	//DSCore.InitPins();
	gpio_init(PIN_DCC0);
	gpio_init(PIN_DCC1);
	gpio_set_dir(PIN_DCC0, true);
	gpio_set_dir(PIN_DCC1, true);
	gpio_disable_pulls(PIN_DCC0);
	gpio_disable_pulls(PIN_DCC1);
	
	
	irq_set_exclusive_handler(PIO0_IRQ_0, dccrecvIrqHandler);
	irq_set_enabled(PIO0_IRQ_0, true);
	irq_set_priority(PIO0_IRQ_0, PICO_DEFAULT_IRQ_PRIORITY);
	irq_clear(PIO0_IRQ_0);	// クリア
	irq_set_enabled(PIO0_IRQ_0, true);	// 許可

	
	// PIOプログラムを初期化
	dcc_pulse_gen_init(gPio, gPio_SM, gPio_PrgAddr, PIN_DCC0);
	
	//Serial.println("START PIO");

}


void PIODCC_ChangeProcess(uint8_t inNextState)
{
	
	switch(state_piodcc)
	{
	case PIODCC_STATE_IDLE:
		
		switch(inNextState)
		{
		case PIODCC_STATE_OUTDCC:
			//PIO動作開始
			PIODCC_StartPIO();
			state_dccgen = PIODCC_GEN_DCCSEND;
			break;
			
		case PIODCC_STATE_OUTCV:
			//PIO動作開始
			PIODCC_StartPIO();
			state_dccgen = PIODCC_GEN_CVSEND;
			
			break;
			
		case PIODCC_STATE_OUTANA:
		case PIODCC_STATE_ERROR:
			//何もしない
			break;
			
		}
		
		
		break;
		
	case PIODCC_STATE_OUTDCC:
		
		switch(inNextState)
		{
		case PIODCC_STATE_IDLE:
		case PIODCC_STATE_OUTANA:
		case PIODCC_STATE_ERROR:
			//PIO動作終了
			PIODCC_StopPIO();
			state_dccgen = PIODCC_GEN_READY;
			break;
		}
		
		break;
		
	case PIODCC_STATE_OUTCV:
		
		switch(inNextState)
		{
		case PIODCC_STATE_IDLE:
		case PIODCC_STATE_OUTANA:
		case PIODCC_STATE_ERROR:
			//PIO動作終了
			PIODCC_StopPIO();
			state_dccgen = PIODCC_GEN_READY;
			break;
		}
		
		break;
		
	case PIODCC_STATE_OUTANA:
		break;
		
	case PIODCC_STATE_ERROR:
		break;
	}
	
	state_piodcc = inNextState;
}


void dccrecvIrqHandler()
{
	int aTimeOutCheckCnt = 1;
	int aFIFOstackCnt = 0;
	uint32_t aBuf = 0xFFFFFFFF;
	word aCVValue = 0;
	
	uint8_t aCVScan_state = DSC_STATE_IDLE;
	
	// clear interrupt first
	pio_interrupt_clear(gPio, 0);
		
		
	switch(state_dccgen)
	{
	case PIODCC_GEN_READY:
		//データなし
		aBuf = 0xFFFFFFFF;
		
		//遷移しない

		//PIO TX FIFO set
		pio_sm_put_blocking(gPio, gPio_SM, aBuf);

		break;
		
	case PIODCC_GEN_DCCSEND:
		//パケット送信
		
		switch(state_bidicutout)
		{
			case 0:

				if (DSCore.IsScanDataLen() > 0)
				{
					aBuf = DSCore.ScanData();
					//Serial.println("GET");
				}
				else
				{
					//バッファなし
					gInterval_BiDi++;

					if( gInterval_BiDi >= 20)
					{
						gInterval_BiDi = 20;
					}

					//
					if( (paramman.GetFlagBidiCutout() == true) &&  (gFlag_PacketType == 1) && (gInterval_BiDi == 20) )
					{
						//BiDi信号送信準備
						state_bidicutout = 1;

						gCounter_BiDi = 0;
						gInterval_BiDi = 0;
						gFlag_PacketType  = 0;

						DSCore.FillPacketBit();
						aBuf = DSCore.ScanData();
					}
					else
					{
						//前回の残りにくっつける
						DSCore.PreparePacketStack();

						//2回送信不要のため、Scanする
						gFlag_PacketType = DSCore.Scan();
						
						//データを取得
						aBuf = DSCore.ScanData();
						//Serial.println("SCAN/GET");	
					}
				}

				//PIO TX FIFO set
				pio_sm_put_blocking(gPio, gPio_SM, aBuf);

				break;

			case 1:

				//PIO TX FIFO set
				pio_sm_put_blocking(gPio, gPio_SM, 0xFFFFFFFF);

				//割り込み不許可
				irq_set_enabled(PIO0_IRQ_0, false);
				
				pio_sm_set_enabled(gPio, gPio_SM, false);

				//PIOからGPIOに戻す
				gpio_init(PIN_DCC_A);
				gpio_init(PIN_DCC_B);
				gpio_set_dir(PIN_DCC_A, GPIO_OUT);
				gpio_set_dir(PIN_DCC_B, GPIO_OUT);
				
				//Low
				gpio_put(PIN_DCC_A, 0);
				gpio_put(PIN_DCC_B, 1);

				//BiDiパルス生成
				add_repeating_timer_us( -10, timer_callback_bidi29us, NULL, &timer_bidi29us_task ); // Timerの起動

				break;
		}


		
		break;

	}
	

}

bool timer_callback_bidi29us(repeating_timer_t *rt)
{
	bool aRet = true;

	if ( (gCounter_BiDi >= 0) && (gCounter_BiDi < 47))
	{
		//出力短絡
		gpio_put(PIN_DCC_A, 1);
		gpio_put(PIN_DCC_B, 1);
	}
	else if ( (gCounter_BiDi >= 47) && (gCounter_BiDi < 49))
	{
		//Low
		gpio_put(PIN_DCC_A, 1);
		gpio_put(PIN_DCC_B, 0);
	}
	else
	{
		//High
		gpio_put(PIN_DCC_A, 0);
		gpio_put(PIN_DCC_B, 1);

		//割り込み終了
		cancel_repeating_timer(&timer_bidi29us_task);
		state_bidicutout = 0;

		//2回送信不要のため、Scanする
		gFlag_PacketType = DSCore.Scan();
		
		//データを取得
		uint32_t aBuf = DSCore.ScanData();		

		//PIOに戻す
		irq_set_exclusive_handler(PIO0_IRQ_0, dccrecvIrqHandler);
		irq_set_enabled(PIO0_IRQ_0, true);
		irq_set_priority(PIO0_IRQ_0, PICO_DEFAULT_IRQ_PRIORITY);
		irq_clear(PIO0_IRQ_0);	// クリア
		irq_set_enabled(PIO0_IRQ_0, true);	// 許可
		
		// PIOプログラムを初期化
		dcc_pulse_gen_init(gPio, gPio_SM, gPio_PrgAddr, PIN_DCC0);

		//PIO TX FIFO set
		pio_sm_put_blocking(gPio, gPio_SM, aBuf);

		//この割り込みを終了
		aRet = false;
	}

	gCounter_BiDi++;

	return aRet;
}

bool timer_callback_100ms(repeating_timer_t *rt)
{
	//VRの定周期取得(LPF 5Hz)
	UI.IntervalVRSpeed();

	//UI関連
	//UI.Process();

	//ボタン初期化・取得
	UI.InitMatrixButton();
	add_repeating_timer_us( -200, timer_callback_buttoninput, NULL, &timer_buttoninput_task ); // Timerの起動



	return true;
}

bool timer_callback_500ms(repeating_timer_t *rt)
{

	/*Sensor check */
	ProcessSensors();

	/*UI slow interval*/
	UI.ProcessInterval();

	return true;
}

bool timer_callback_buttoninput(repeating_timer_t *rt)
{
	bool aRet = UI.IntervalMatrixButton();

	if( aRet == false)
	{
		cancel_repeating_timer(&timer_buttoninput_task);
	}

	return aRet;
}


/**
 * @brief DSairコマンドをセットし実行
 * 
 * @param inFunction コマンド名
 * @param inArg1 引数1
 * @param inArg2 引数2
 * @param inArg3 引数3
 * @param inArg4 引数4
 */
void callback_control(const String inFunction, const word inArg1, const word inArg2, const word inArg3, const word inArg4)
{
	function = inFunction;
	arguments[0] = inArg1;
	arguments[1] = inArg2;
	arguments[2] = inArg3;
	arguments[3] = inArg4;
	
	//処理実行
	dispatch(1);
}

/**
 * @brief UIへの反映処理
 * 
 */
void ApplyCommandUI()
{
	
	//UIへの反映処理
	if ((function == "setLocoDirection") || ((function == "DI")))
	{
		//UIへの反映
		UI.NotifyDCCDirection(arguments[0], arguments[1]);
		
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		//UIへの反映
		UI.NotifyDCCPower(arguments[0]);
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		//UIへの反映
		UI.NotifyDCCSpeed(arguments[0], arguments[1]);
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		
	}
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		
	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		
	}
	
	
	
}

/**
 * @brief Set the Analog Speed to PWM output function
 * 
 * @param inSpeed 速度[0-1023]
 * @param inDir 進行方向[1: FWD, 2: REV]
 */

void SetAnalogSpeed(word inSpeed, uint8_t inDir)
{
	
	word aSpeed = inSpeed >> 2;
	
	if( aSpeed > 255)
	{
		aSpeed = 255;
	}
	
	if( inSpeed == 0)
	{
		FreeAnalogSpeed();
	}
	else
	{
		
		analogWriteFreq(32000);
		
		if( inDir == 2)
		{
			//REV
			analogWrite(PIN_PWMA, 255 - aSpeed);
			analogWrite(PIN_PWMB, 255);
		}
		else
		{
			//FWD
			analogWrite(PIN_PWMB, 255 - aSpeed);
			analogWrite(PIN_PWMA, 255);
		}
	}
}

/**
 * @brief Deactivate Analog PWM poutput
 * 
 */
void FreeAnalogSpeed()
{
	analogWrite(PIN_PWMB, 0);
	analogWrite(PIN_PWMA, 0);
}


//-----------------------------------------------------------------------
// WiFi関連処理


void setup_wifi() {

  // check for the WiFi module:
  if (WiFi.status() == WL_NO_MODULE) {
	Serial.println("Communication with WiFi module failed!");
    // don't continue
    return;
  }

  String fv = WiFi.firmwareVersion();

  if (fv < WIFI_FIRMWARE_LATEST_VERSION) {
    Serial.println("Please upgrade the firmware");
  }

  //WiFi.mode(WIFI_STA); // 20231229 HMX
  //WiFi.setHostname("flashair"); // 20231229 HMX

  // by default the local IP address will be 192.168.42.1
  // you can override it with the following:
  // WiFi.config(IPAddress(192, 168, 42, 1));

  WiFi.noLowPowerMode();

  IPAddress ipAdr = IPAddress(192, 168, 42, 1);      // 20231229 HMX
  IPAddress ipSubnet = IPAddress(255, 255, 255, 0);  // 20231229 HMX
  WiFi.softAPConfig(ipAdr, ipAdr, ipSubnet);         // 20231229 HMX

  // print the network name (SSID);
  Serial.print("Creating access point named: ");
  Serial.println(ssid);

  // Create open network. Change this line if you want to create an WEP network:
  // status = WiFi.beginAP(ssid, pass);// 20231229 HMX
  status = WiFi.softAP(ssid, pass);  // 20231229 HMX

  if (status != WL_AP_LISTENING) {
    Serial.println("Creating access point failed");
    // don't continue
    return;
  }
}

void setup_server() {

  // start the web server on port 80
  server.begin();
}

void loop_wifi() 
{
	uint8_t aReply264;
	
	WiFiClient client_new = server.accept();  // listen for incoming clients

	UI.WiFiStatus = WiFi.status() == WL_CONNECTED ? 1 : 0;

	if (client_new)
	{
		for (int i = 0; i < 4; i++)
		{
			if (!clients[i]) 
			{
				clients[i] = client_new;
				break;
			}
		}
	}


	//Communication with client
	WiFiClient client = clients[index_client];
	index_client++;

	if( index_client > 3)
	{
		index_client = 0;
	}

	// if you get a client,
	//Serial.println("new client");  // print a message out the serial port
	String currentLine = "";       // make a String to hold incoming data from the client

	unsigned long aTimeout_base = millis();
	aReply264 = 0;
		
	while (client && client.available() > 0)
	{
		if (millis() - aTimeout_base > 50)
		{
			break;
		}
				// loop while the client's connected
		// if (client.available())
		if (client.available() != 0) // HMX 20240113a 
		{
			//Serial.print("*");
			// if there's bytes to read from the client,
			char c = client.read();  // read a byte, then
				
			//Serial.write(c);	// print it out the serial monitor
			if (c == '\n')
			{
				//HTTPのコマンドを1行ずつ処理する
				if (currentLine.length() == 0)
				{
					//すべてのコマンドを受信したら応答して終了
					
					if( aReply264 == 1)
					{
						
						replyHTML264(client);

						//ステータス取得するスロットルが存在すること通知
						UI.NotifyWiFiClient();
						
					}
					else if( aReply264 == 2)
					{
						replyWebApp(client);
						client.stop();
					}
					else
					{
						reply200Ok(client);
					}
					
					// break out of the while loop:
					break;
				}
				else
				{
					// FlashAir互換コマンドを受け付ける
					if (currentLine.indexOf("GET /command.cgi?op=131&ADDR=0&LEN=64&DATA=") == 0) 
					{
						//Extract DSair WiFi Command
						int aStart = 43;  //currentLine.indexOf("&DATA=");
						int aEnd = currentLine.indexOf(")");
						String aCommand = currentLine.substring(aStart, aEnd + 1);

						//Core 0と共用変数の処理
						gRequest_wifi = aCommand;

						//replyHTML(client);

						//コマンドを受け付けたらあとの処理は省略する
						aReply264 = 1;

						//break;
					}
					else if (currentLine.indexOf("GET /command.cgi?op=130&ADDR=128&LEN=264") == 0) 
					{
						// FlashAir互換・ステータス応答処理
						aReply264 = 1;

						//replyHTML264(client);

						//ステータス取得するスロットルが存在すること通知
						//UI.NotifyWiFiClient();

						//コマンドを受け付けたらあとの処理は省略する
						//break;
					}
					else if (currentLine.indexOf("GET / ") == 0) 
					{
						aReply264 = 2;
					}
					else if (currentLine.indexOf("GET /index.html") == 0) 
					{
						aReply264 = 2;

					}					

					currentLine = "";
				}
			}
			else if (c != '\r')
			{
				currentLine += c;  // add it to the end of the currentLine
			}
		} 
		else 
		{
			//Nothing to do 
		}
	}
	

	// stop any clients which disconnect
	for (int i = 0; i < 4; i++)
	{
		if (clients[i] && !clients[i].connected())
		{
			clients[i].flush();			
			clients[i].stop();
		}
	}

}

void replyHTML264(WiFiClient client)
{
	client.println("HTTP/1.1 200 OK");
	//client.println("Server: Xcerebellum/1.0");
	client.println("Content-Length: 264");
	client.println("Connection: close");
	client.print("Content-Type: application/octet-stream\r\n\r\n");
	//client.println();
	
	client.write( (uint8_t *)buffer_sharedStatusMem, 264);
	client.println();
	client.println();
	
}